"""Normalize text using unicode properties."""

import sys
import argparse

from tqdm import tqdm

from icu_tokenizer.normalizer import Normalizer
from icu_tokenizer.utils import TextFileType

CACHE = {}


def add_options(parser: argparse.ArgumentParser):
    """Add options to a parser."""
    parser.add_argument(
        '-l', '--lang', type=str, default='en',
        help='Language identifier')
    parser.add_argument(
        '-p', '--norm-puncts', action='store_true',
        help='Normalize punctuations')
    parser.add_argument(
        '-lc', '--lowercase', action='store_true',
        help='Cast all characters to lowercase')

    parser.add_argument(
        '-i', '--inputs', type=TextFileType('r'),
        nargs='+', default=[sys.stdin],
        help='Input files. Defaults to stdin.')
    parser.add_argument(
        '-o', '--output', type=TextFileType('w'), default=sys.stdout,
        help='Output file. Defaults to stdout.')

    parser.add_argument(
        '-j', '--num-workers', type=int, default=0,
        help='Number of processes to use')
    parser.add_argument(
        '--show-pbar', action='store_true',
        help='Show progressbar')


def main(args: argparse.Namespace):  # noqa
    if args.num_workers == 0:
        import multiprocessing.dummy as multiprocessing
        args.num_workers = 1
    else:
        import multiprocessing

    if args.num_workers < 0:  # Use all cores
        args.num_workers = multiprocessing.cpu_count()

    def create_chunk_input_stream():
        chunk = []
        for f in args.inputs:
            for line in f:
                chunk.append(line)
                if len(chunk) >= 256:
                    yield chunk
                    chunk = []
        if len(chunk) > 0:
            yield chunk

    pbar = None
    if args.show_pbar:
        pbar = tqdm()

    with multiprocessing.Pool(
        args.num_workers,
        initializer=worker_init_fn,
        initargs=[args.lang, args.norm_puncts, args.lowercase]
    ) as pool:
        for chunk in pool.imap(worker_fn, create_chunk_input_stream()):
            if pbar is not None:
                pbar.update(len(chunk))
            for line in chunk:
                args.output.write(line + '\n')
    args.output.flush()

    if pbar is not None:
        pbar.close()


def worker_init_fn(lang: str, norm_puncts: bool, lowercase: bool):  # noqa
    CACHE['normalizer'] = Normalizer(lang, norm_puncts)
    CACHE['lowercase'] = lowercase


def worker_fn(texts):  # noqa
    normalize_fn = CACHE['normalizer'].normalize
    texts = [normalize_fn(t) for t in texts]
    if CACHE['lowercase']:
        texts = [t.lower() for t in texts]
    return texts



from typing import List

from icu import BreakIterator, Locale

from icu_tokenizer.utils import apply_break_iterator


class SentSplitter(object):
    """ICU sentence splitter.

    Usage:

    >>> splitter = SentSplitter(lang)
    >>> sents: List[str] = splitter.split(paragraph)
    """

    def __init__(self, lang: str = 'en'):
        """SentSplitter."""
        self.lang = lang
        self.locale = Locale(lang)
        self.break_iterator = \
            BreakIterator.createSentenceInstance(self.locale)

    def split(self, text: str) -> List[str]:
        """Split a sentence with the ICU sentence splitter."""
        return apply_break_iterator(self.break_iterator, text)



# -*- coding: utf-8 -*-

import argparse
import re
import sys
from typing import List

import icu
import regex
from icu import BreakIterator


def apply_break_iterator(
    break_iterator: BreakIterator,
    text: str
) -> List[str]:
    """Apply ICU break iterator on a text."""
    break_iterator.setText(text)
    parts = []
    p0 = 0
    for p1 in break_iterator:
        part = text[p0:p1].strip()
        if len(part) > 0:
            parts.append(part)
        p0 = p1
    return parts


def get_all_unicode_chars():
    """Get all unicode characters."""
    all_unicode_chars = []
    i = 0
    while True:
        try:
            all_unicode_chars.append(chr(i))
        except ValueError:
            break
        i += 1
    return all_unicode_chars


def get_versions() -> dict:
    """Get versions of the various dependecies related to icu_tokenizer."""
    versions = {
        'icu': icu.ICU_VERSION,
        'PyICU': icu.VERSION,
        'regex': regex.__version__
    }

    try:
        import opencc
        versions['opencc'] = opencc.__version__
    except ImportError:
        pass

    return versions


class TextFileType(argparse.FileType):
    """argparse.FileType modified for utf-8 text files."""

    def __init__(self, mode: str = 'r', bufsize: int = -1):
        """TextFileType."""
        self._mode = mode
        self._bufsize = bufsize
        self._encoding = 'utf-8'
        self._errors = 'ignore'

    def __call__(self, string):  # noqa
        # the special argument "-" means sys.std{in,out}
        if string == '-':
            if 'r' in self._mode:
                return sys.stdin
            elif 'w' in self._mode:
                return sys.stdout
            else:
                msg = 'argument "-" with mode {}'.format(self._mode)
                raise ValueError(msg)

        # all other arguments are used as file names
        try:
            return open(string, self._mode, self._bufsize, self._encoding,
                        self._errors)
        except OSError as e:
            msg = "can't open '{}': {}".format(string, e)
            raise argparse.ArgumentTypeError(msg)



from icu_tokenizer.normalizer import Normalizer
from icu_tokenizer.sent_splitter import SentSplitter
from icu_tokenizer.tokenizer import Tokenizer

__all__ = ['Normalizer', 'SentSplitter', 'Tokenizer']



# -*- coding: utf-8 -*-

import re
import unicodedata
from typing import Dict, List

import regex

from icu_tokenizer.utils import get_all_unicode_chars


class Normalizer(object):
    """Unicode information based normalizer.

    Does the following

    - Ensure NFKC format
    - Handle pseudo-spaces (for numbers)
    - Normalize by unicode categories \
      https://www.fileformat.info/info/unicode/category/index.htm

        - ``[C*|So|Z*]`` → ' '
        - ``[Pc]`` → ``_``
        - ``[Pd]`` → ``-``
        - ``[Pf|Pi]`` → ``"``  (except for ``'``)
        - ``[Ps]`` → ``(``  (except for ``{``, ``[``)
        - ``[Pe]`` → ``)``  (except for ``}``, ``]``)
    - Normalize Nd (Numbers)
    - Account for some outliers
    - Remove non printable characters
    - Normalize whitespace characters
    - Perform language specific normalization

    Usage:

    >>> normalizer = Normalizer(lang, norm_puncts=True)
    >>> norm_text: str = normalizer.normalize(text)
    """

    def __init__(self, lang: str = 'en', norm_puncts: bool = False):
        """Normalizer.

        Args:
            lang (str, optional): Language identifier. Defaults to 'en'.
            norm_puncts (bool, optional): Normalize punctuations?.
                Defaults to False.
        """
        # Handle control tokens
        self.ignore_pattern = regex.compile(r'\p{C}|\p{So}|\p{Z}')

        # Handle pseudo-spaces
        # Random note: it appears pseudo-spaces primarily makes a difference
        # when numbers are involved
        self.pseudo_num_pattern = re.compile(r'(\d) (\d)')

        # Punctuation and number replace maps
        self.num_pattern = regex.compile(r'\p{Nd}+')
        self.punct_replace_map = self.punct_pattern = None
        if norm_puncts:
            self.punct_replace_map = make_punct_replace_map()
            self.punct_pattern = \
                make_pattern_from_keys(self.punct_replace_map.keys())

        # Other language specific normalizers
        lang_replace_map = make_lang_specific_replace_map(lang)
        self.lang_replace_map = self.lang_replace_pattern = None
        if len(lang_replace_map) > 0:
            self.lang_replace_map = lang_replace_map
            self.lang_replace_pattern = \
                make_pattern_from_keys(lang_replace_map.keys())

    def _num_replace_fn(self, match: re.Match) -> str:
        return str(int(match.group(0)))

    def _punct_replace_fn(self, match: re.Match) -> str:
        return self.punct_replace_map[match.group(0)]

    def _lang_replace_fn(self, match: re.Match) -> str:
        return self.lang_replace_map[match.group(0)]

    def normalize(self, text: str) -> str:
        """Perform normalization.

        Args:
            text (str): Input text

        Returns:
            str: Normalized text
        """
        text = unicodedata.normalize('NFKC', text)

        text = self.pseudo_num_pattern.sub(r'\1.\2', text)
        text = self.num_pattern.sub(self._num_replace_fn, text)
        if self.punct_pattern is not None:
            text = self.punct_pattern.sub(self._punct_replace_fn, text)

        text = self.ignore_pattern.sub(' ', text)
        text = ' '.join(text.split())  # Normalize whitespace

        if self.lang_replace_pattern is not None:
            text = self.lang_replace_pattern(self._lang_replace_fn, text)

        return text


def make_pattern_from_keys(keys: List[str]) -> re.Pattern:
    """Make a re.Pattern that matches a list of strings."""
    keys = sorted(keys, key=lambda x: len(x), reverse=True)
    pattern_str = r'|'.join(re.escape(k) for k in keys)
    return re.compile(pattern_str)


def make_punct_replace_map() -> Dict[str, str]:
    """Make the punctuation replace map."""
    # Generate punctuation and number replace maps
    punct_replace_map = {}

    # Normalization rules based on unicode category
    punct_exceptions = {"'", '[', ']', '{', '}'}
    for c in get_all_unicode_chars():
        if c in punct_exceptions:
            continue

        cat = unicodedata.category(c)
        if cat == 'Pc':
            punct_replace_map[c] = '_'
        elif cat == 'Pd':
            punct_replace_map[c] = '-'
        elif cat == 'Pe':
            punct_replace_map[c] = ')'
        elif cat == 'Pf':
            punct_replace_map[c] = '"'
        elif cat == 'Pi':
            punct_replace_map[c] = '"'
        elif cat == 'Ps':
            punct_replace_map[c] = '('

    # User provided rules

    # Soft hyphen
    punct_replace_map['\xad'] = ''

    # Double quotes
    punct_replace_map["''"] = '"'
    punct_replace_map["´´"] = '"'
    punct_replace_map['„'] = '"'

    # Apostrophes
    punct_replace_map["`"] = "'"
    punct_replace_map['´'] = "'"
    punct_replace_map['‘'] = "'"
    punct_replace_map['’'] = "'"
    punct_replace_map['‚'] = "'"  # Not a comma

    # Brackets
    punct_replace_map['【'] = '['
    punct_replace_map['】'] = ']'
    punct_replace_map['［'] = '['
    punct_replace_map['］'] = ']'

    # Common unicode variations
    punct_replace_map['∶'] = ':'
    punct_replace_map['？'] = '?'
    punct_replace_map['．'] = '.'
    punct_replace_map['━'] = '-'
    punct_replace_map['％'] = '%'

    # Chinese punctuations
    punct_replace_map['！'] = '!'
    punct_replace_map['、'] = ','
    punct_replace_map['｜'] = '|'
    punct_replace_map['：'] = ':'
    punct_replace_map['；'] = ';'
    punct_replace_map['，'] = ','
    punct_replace_map['。'] = '.'
    punct_replace_map['～'] = '~'

    # Others
    punct_replace_map['…'] = '...'

    return punct_replace_map


def make_lang_specific_replace_map(lang: str = 'en') -> Dict[str, str]:
    """Create a language specific replace map."""
    replace_map = {}

    if lang == 'ro':
        # Remove diacritics for romanian
        replace_map['Ş'] = 'S'
        replace_map['ş'] = 's'

        replace_map['Ș'] = 'S'
        replace_map['ș'] = 's'

        replace_map['Ţ'] = 'T'
        replace_map['ţ'] = 't'

        replace_map['Ț'] = 'T'
        replace_map['ț'] = 't'

        replace_map['Ă'] = 'A'
        replace_map['ă'] = 'a'

        replace_map['Â'] = 'A'
        replace_map['â'] = 'a'

        replace_map['Î'] = 'I'
        replace_map['î'] = 'i'

    return replace_map



import re
import regex

__all__ = ['email_pattern', 'grubber_url_matcher']


sub_domain_pstr = r'[0-9A-Za-z\-\_\~]+'
top_domain_pstr = r'(?:[.](?:{}))'.format(r'|'.join(re.escape(s) for s in [
    'com', 'org', 'net', 'int', 'edu', 'gov', 'mil', 'ac', 'ad',
    'ae', 'af', 'ag', 'ai', 'al', 'am', 'ao', 'aq', 'ar', 'as',
    'at', 'au', 'aw', 'ax', 'az',
    'ba', 'bb', 'bd', 'be', 'bf', 'bg', 'bh', 'bi', 'bj', 'bm',
    'bn', 'bo', 'br', 'bs', 'bt', 'bw', 'by', 'bz',
    'ca', 'cc', 'cd', 'cf', 'cg', 'ch', 'ci', 'ck', 'cl', 'cm',
    'cn', 'co', 'cr', 'cu', 'cv', 'cw', 'cx', 'cy', 'cz',
    'de', 'dj', 'dk', 'dm', 'do', 'dz',
    'ec', 'ee', 'eg', 'er', 'es', 'et', 'eu',
    'fi', 'fj', 'fk', 'fm', 'fo', 'fr',
    'ga', 'gd', 'ge', 'gf', 'gg', 'gh', 'gi', 'gl', 'gm', 'gn',
    'gp', 'gq', 'gr', 'gs', 'gt', 'gu', 'gw', 'gy',
    'hk', 'hm', 'hn', 'hr', 'ht', 'hu',
    'id', 'ie', 'il', 'im', 'in', 'io', 'iq', 'ir', 'is', 'it',
    'je', 'jm', 'jo', 'jp',
    'ke', 'kg', 'kh', 'ki', 'km', 'kn', 'kp', 'kr', 'kw', 'ky',
    'kz',
    'la', 'lb', 'lc', 'li', 'lk', 'lr', 'ls', 'lt', 'lu', 'lv',
    'ly',
    'ma', 'mc', 'md', 'me', 'mg', 'mh', 'mk', 'ml', 'mm', 'mn',
    'mo', 'mp', 'mq', 'mr', 'ms', 'mt', 'mu', 'mv', 'mw', 'mx',
    'my', 'mz',
    'na', 'nc', 'ne', 'nf', 'ng', 'ni', 'nl', 'no', 'np', 'nr',
    'nu', 'nz',
    'om',
    'pa', 'pe', 'pf', 'pg', 'ph', 'pk', 'pl', 'pm', 'pn', 'pr',
    'ps', 'pt', 'pw', 'py',
    'qa',
    're', 'ro', 'rs', 'ru', 'rw',
    'sa', 'sb', 'sc', 'sd', 'se', 'sg', 'sh', 'si', 'sk', 'sl',
    'sm', 'sn', 'so', 'sr', 'ss', 'st', 'su', 'sv', 'sx', 'sy',
    'sz',
    'tc', 'td', 'tf', 'tg', 'th', 'tj', 'tk', 'tl', 'tm', 'tn',
    'to', 'tr', 'tt', 'tv', 'tw', 'tz',
    'ua', 'ug', 'uk', 'us', 'uy', 'uz',
    'va', 'vc', 've', 'vg', 'vi', 'vn', 'vu',
    'wf', 'ws',
    'ye', 'yt',
    'za', 'zm', 'zw'
]))  # https://en.wikipedia.org/wiki/List_of_Internet_top-level_domains

domain_pstr = r'(?:{sub_domain}\.)*{sub_domain}{top_domain}'.format(
    sub_domain=sub_domain_pstr,
    top_domain=top_domain_pstr,
)

# https://stackoverflow.com/questions/2049502/what-characters-are-allowed-in-an-email-address
local_part_valid_word = r'[0-9A-Za-z\!\#\$\%\&\'\*\+\-\/\=\?\^\_\`\{\|\}\~]+'
local_part_pstr = r'{word}(?:\.{word})*'.format(word=local_part_valid_word)

email_pstr = r'({})'.format(
    r'{local_part}\@{domain}'.format(
        local_part=local_part_pstr,
        domain=domain_pstr,
    )
)
email_pattern: regex.regex.Pattern = regex.compile(email_pstr, re.IGNORECASE)
"""Custom email matcher based on
https://en.wikipedia.org/wiki/International_email
"""


# A customized grubber v1 URL matcher
# Designed to work with urls starting with https, http, ftp, or www
grubber_url_pstr = r'(?i)\b((?:(?:https|http|ftp):/{1,3}|www[.])[^\s()<>\（\）\【\】]+(?:\([\w\d]+\)|(?:[^!"#$%&\'()*+,\-./:;<=>?@\[\]\s\（\）\【\】。，？！]|/)))'  # noqa
grubber_url_matcher: re.Pattern = re.compile(grubber_url_pstr, re.ASCII)
"""Grubber v1 URL matcher with additional rules to account for chinese
punctuations.

Designed to work with urls starting with https, http, ftp, or www.
"""



import argparse
from types import ModuleType
from typing import Dict

from icu_tokenizer.bin import normalize as normalize_module
from icu_tokenizer.bin import split as split_module
from icu_tokenizer.bin import tokenize as tokenize_module

SUBCOMMANDS: Dict[str, ModuleType] = {
    'normalize': normalize_module,
    'split': split_module,
    'tokenize': tokenize_module,
}


def make_parser() -> argparse.ArgumentParser:
    """Make the parser for the main program."""
    parser = argparse.ArgumentParser(
        prog='python3 -m mt_experiments',
        description='Machine Translation Experimentation Toolkit'
    )
    subparsers = parser.add_subparsers(dest='subcommand')

    for subcommand, module in SUBCOMMANDS.items():
        module.add_options(subparsers.add_parser(
            subcommand, help=module.__doc__,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
        ))

    return parser


if __name__ == "__main__":
    parser = make_parser()
    args = parser.parse_args()

    if args.subcommand in SUBCOMMANDS:
        module = SUBCOMMANDS[args.subcommand]
        module.main(args)
    else:
        parser.print_help()



import re
from typing import List, Union

from icu import BreakIterator, Locale

from icu_tokenizer.url_utils import email_pattern, grubber_url_matcher
from icu_tokenizer.utils import apply_break_iterator

PROTECTED_TEMPLATE = '__PROTECTED_SEQUENCE_{}__'


class Tokenizer(object):
    """ICU based tokenizer with additional functionality to protect sequences.

    Usage:

    >>> tokenizer = Tokenizer(
            lang,
            annotate_hyphens: bool,
            protect_emails_urls: bool,
            extra_protected_patterns: List[Union[str, re.Pattern]] = [],
        )
    >>> tokens: List[str] = tokenizer.tokenize(text)
    """

    HYPHEN_PATTERN = re.compile(r'(\w)\-(?=\w)')
    HYPHEN_PATTERN_REPL = r'\1 @-@ '
    PROTECTED_HYPHEN_PATTERN = re.compile(r'@\-@')

    def __init__(
        self,
        lang: str = 'en',
        annotate_hyphens: bool = False,
        protect_emails_urls: bool = False,
        extra_protected_patterns: List[Union[str, re.Pattern]] = [],
    ):
        """Tokenizer.

        Keyword Arguments:
            lang {str} -- language identifier (default: {'en'})
            annotate_hyphens {bool} -- Protect dashes (default: {False})
            protect_emails_urls {bool} -- Protect urls (default: {False})
            extra_protected_patterns {List[Union[str, re.Pattern]]} --
                A list of regex patterns (default: {[]})
        """
        self.lang = lang
        self.locale = Locale(lang)
        self.break_iterator = \
            BreakIterator.createWordInstance(self.locale)
        self.protected_patterns = []

        self.annotate_hyphens = annotate_hyphens
        if self.annotate_hyphens:
            self.protected_patterns.append(self.PROTECTED_HYPHEN_PATTERN)

        if protect_emails_urls:
            self.protected_patterns.append(email_pattern)
            self.protected_patterns.append(grubber_url_matcher)

        for pattern in extra_protected_patterns:
            if isinstance(pattern, str):
                pattern = re.compile(pattern)
            self.protected_patterns.append(pattern)

    def tokenize(self, text: str) -> List[str]:
        """Tokenize text into list of tokens.

        Args:
            text (str): Raw input text.

        Returns:
            List[str]: List of tokens.
        """
        if self.annotate_hyphens:
            text = self.HYPHEN_PATTERN.sub(self.HYPHEN_PATTERN_REPL, text)

        protected_map = {}

        def protect_replace(match):
            protected_str = PROTECTED_TEMPLATE.format(len(protected_map))
            protected_map[protected_str] = match.group(0)
            return ' {} '.format(protected_str)

        for i, pattern in enumerate(self.protected_patterns):
            text = pattern.sub(protect_replace, text)

        tokens = apply_break_iterator(self.break_iterator, text)
        return [protected_map.get(t, t) for t in tokens]



"""Split lines containing multiple sentences."""

import sys
import argparse

from tqdm import tqdm

from icu_tokenizer.sent_splitter import SentSplitter
from icu_tokenizer.utils import TextFileType

CACHE = {}


def add_options(parser: argparse.ArgumentParser):
    """Add options to a parser."""
    parser.add_argument(
        '-l', '--lang', type=str, default='en',
        help='Language identifier')

    parser.add_argument(
        '-i', '--inputs', type=TextFileType('r'),
        nargs='+', default=[sys.stdin],
        help='Input files. Defaults to stdin.')
    parser.add_argument(
        '-o', '--output', type=TextFileType('w'), default=sys.stdout,
        help='Output file. Defaults to stdout.')

    parser.add_argument(
        '-j', '--num-workers', type=int, default=0,
        help='Number of processes to use')
    parser.add_argument(
        '--show-pbar', action='store_true',
        help='Show progressbar')
    parser.add_argument(
        '--verbose', action='store_true',
        help='Print splits to stderr')


def main(args: argparse.Namespace):  # noqa
    if args.num_workers == 0:
        import multiprocessing.dummy as multiprocessing
        args.num_workers = 1
    else:
        import multiprocessing

    if args.num_workers < 0:  # Use all cores
        args.num_workers = multiprocessing.cpu_count()

    def create_chunk_input_stream():
        chunk = []
        for f in args.inputs:
            for line in f:
                chunk.append(line)
                if len(chunk) >= 256:
                    yield chunk
                    chunk = []
        if len(chunk) > 0:
            yield chunk

    pbar = None
    if args.show_pbar:
        pbar = tqdm()

    with multiprocessing.Pool(
        args.num_workers,
        initializer=worker_init_fn,
        initargs=[args.lang]
    ) as pool:
        for chunk in pool.imap(worker_fn, create_chunk_input_stream()):
            if pbar is not None:
                pbar.update(len(chunk))
            for sents in chunk:
                for sent in sents:
                    args.output.write(sent + '\n')
                if args.verbose and len(sents) > 1:
                    sys.stderr.write('\rSplitting done: {}\n'.format(sents))
    sys.stderr.flush()
    args.output.flush()

    if pbar is not None:
        pbar.close()


def worker_init_fn(lang: str):  # noqa
    CACHE['sent_splitter'] = SentSplitter(lang)


def worker_fn(texts):  # noqa
    split_fn = CACHE['sent_splitter'].split
    return [split_fn(t) for t in texts]



import setuptools


def get_long_description():
    """Reads and return README as a string."""
    with open('README.md', 'rb') as f:
        return f.read().decode('utf-8', errors='ignore')


with open('requirements.txt', 'r') as f:
    install_requires = f.read()


setuptools.setup(
    name='icu_tokenizer',
    version='0.0.1',
    author='Wang Ming Rui',
    author_email='mingruimingrui@hotmail.com',
    description="ICU based universal language tokenizer",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/mingruimingrui/ICU-tokenizer",

    install_requires=install_requires,
    packages=['icu_tokenizer', 'icu_tokenizer.bin'],

    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Localization',
    ],
    license='MIT License'
)
